import numpy
import rospy
import math

from std_msgs.msg import String, Empty
from sensor_msgs.msg import Image, Range, NavSatFix, MagneticField

from webots_ros.msg import Float64Stamped
from webots_ros.srv import sensor_enable, motor_set_velocity, motor_set_position

isInitialized = False
isCameraEnabled = False
isRangeCameraEnabled = False
isSonarEnabled = False
isEncoderEnabled = False

controllerName = None
TIME_STEP = 32

distance = [0] * 16
ground_truth = [0, 0, 0]
odometry = [None, None, None]

_lastLeftEncoder = None
_lastLeftEncoderStamp = None
_lastRightEncoder = None
_lastRightEncoderStamp = None

_leftSpeed = 0
_rightSpeed = 0

def move(ls,rs):
    """ Make the robot move.

    Args:
        ls (float): left wheel speed in rad/s
        rs (float): right wheel speed in rad/s
    """
    global isInitialized
    if isInitialized:
        global lwv
        global rwv
        max_speed = 5.24
        if abs(ls) > max_speed:
            rs = rs * max_speed / abs(ls)
            ls = ls * max_speed / abs(ls)
        if abs(rs) > max_speed:
            ls = ls * max_speed / abs(rs)
            rs = rs * max_speed / abs(rs)
        lwv(ls)
        rwv(rs)
        rospy.sleep(0.032)

def sleep(t):
    """ Program pause (the robot keeps moving if the speeds are not zero).

    Args:
        t (float): seconds
    """
    rospy.sleep(t)
    
def stop():
    """ Stop the robot.

    Args:
    """
    move(0,0)

def tilt(a):
    """ Tilt the Kinect.

    Args:
        a (float): radians
    """
    global isInitialized
    if isInitialized:
        global tmv
        global tmp
        tmv(1.0)
        if a>0.47:
            a = 0.47
        if a<-0.47:
            a = -0.47
        tmp(a)
    
def _nameCallback(data):
    global controllerName
    controllerName = data.data

def enableMotor(w,v_ini,p_ini):
    vp = rospy.ServiceProxy(controllerName+'/'+w+'/set_velocity',motor_set_velocity)
    pp = rospy.ServiceProxy(controllerName+'/'+w+'/set_position',motor_set_position)
    vp(v_ini)
    pp(p_ini)
    return vp, pp

def enableEncoder(deviceName,cb):
    sp = rospy.ServiceProxy(controllerName+'/'+deviceName+'/enable',sensor_enable)
    sp(TIME_STEP)
    sb = rospy.Subscriber(controllerName+'/'+deviceName+'/value',Float64Stamped,cb)

def _leftEncoder(data):
    global leftEncoder
    leftEncoder = data.data
    global _lastLeftEncoder
    global _lastLeftEncoderStamp
    if not _lastLeftEncoder is None:
        dt = data.header.stamp - _lastLeftEncoderStamp
        global _leftSpeed
        r = 0.1953 / 2    # wheel radius
        _leftSpeed = r * (leftEncoder - _lastLeftEncoder) / dt.to_sec()
    _lastLeftEncoder = leftEncoder
    _lastLeftEncoderStamp = data.header.stamp

def _rightEncoder(data):
    global rightEncoder
    global enc_pbl
    rightEncoder = data.data
    global _lastRightEncoder
    global _lastRightEncoderStamp
    if not _lastRightEncoder is None:
        dt = data.header.stamp - _lastRightEncoderStamp
        global _rightSpeed
        r = 0.1953 / 2    # wheel radius
        _rightSpeed = r * (rightEncoder - _lastRightEncoder) / dt.to_sec()
        global _w
        global _R
        _w = (_rightSpeed - _leftSpeed) / 0.33 # distance between the two wheels
        try:
            _R = 0.33 / 2 * (_rightSpeed + _leftSpeed) / (_rightSpeed - _leftSpeed)
        except ZeroDivisionError:
            _R = float('inf')
        global odometry
        if not odometry[0] is None and not odometry[2] is None:
            c = math.cos(odometry[2])
            s = math.sin(odometry[2])
            if _R == float('inf'):
                v = _rightSpeed
                odometry[0] = odometry[0] + v * c * dt.to_sec()
                odometry[1] = odometry[1] + v * s * dt.to_sec()
            elif _R == 0:
                odometry[2] = angleNormalize(odometry[2]+_w*dt.to_sec())
            else:
                ICCx = odometry[0] - _R * s
                ICCy = odometry[1] + _R * c
                c = math.cos(_w*dt.to_sec())
                s = math.sin(_w*dt.to_sec())
                x = odometry[0]
                y = odometry[1]
                nx = c*(x-ICCx)-s*(y-ICCy)+ICCx
                ny = s*(x-ICCx)+c*(y-ICCy)+ICCy
                odometry[0] = nx
                odometry[1] = ny
                odometry[2] = angleNormalize(odometry[2]+_w*dt.to_sec())
    _lastRightEncoder = rightEncoder
    _lastRightEncoderStamp = data.header.stamp
    enc_pbl.publish(Empty())
    
def angleNormalize(a):
    while a > math.pi:
        a = a - 2*math.pi
    while a < -math.pi:
        a = a + 2*math.pi
    return a

def _gpsCallback(data):
    global ground_truth
    ground_truth[0] = data.longitude
    ground_truth[1] = data.latitude
    global odometry
    if odometry[0] is None:
        odometry[0] = ground_truth[0]
        odometry[1] = ground_truth[1]
        
def _compassCallback(data):
    global ground_truth
    mf = data.magnetic_field
    bearing = math.atan2(mf.x, mf.z)
    bearing = -bearing - math.pi/2
    if bearing > math.pi:
        bearing = bearing - 2*math.pi
    elif bearing < -math.pi:
        bearing = bearing + 2*math.pi
    ground_truth[2] = bearing
    global odometry
    if odometry[2] is None:
        odometry[2] = ground_truth[2]
    pose_pbl.publish(Empty())
    
def enablePose():
    sp = rospy.ServiceProxy(controllerName+'/gps/enable',sensor_enable)
    sp(TIME_STEP)
    sb = rospy.Subscriber(controllerName+'/gps/values',NavSatFix,_gpsCallback)

    sp = rospy.ServiceProxy(controllerName+'/compass/enable',sensor_enable)
    sp(TIME_STEP)
    sb = rospy.Subscriber(controllerName+'/compass/values',MagneticField,_compassCallback)
    
def init():
    global isInitialized
    if not isInitialized:
        rospy.init_node('controller', anonymous=True)
        isInitialized = True
        global controllerName
        sb = rospy.Subscriber("/model_name", String, _nameCallback)
        while controllerName is None:
            rospy.sleep(0.1)
        sb.unregister()
        global lwv
        global rwv
        lwv, lwp = enableMotor('left_wheel',0.0,float('inf'))
        rwv, rwp = enableMotor('right_wheel',0.0,float('inf'))
        global tmv
        global tmp
        tmv, tmp = enableMotor('tilt_motor',0.0,0.0)
        enableCamera()
        enableSonar()
        enableEncoder('left_wheel_sensor',_leftEncoder)
        enableEncoder('right_wheel_sensor',_rightEncoder)
        global enc_pbl
        enc_pbl = rospy.Publisher(controllerName+'/encoder_updated',Empty)
        enablePose() # GPS and compass
        global pose_pbl
        pose_pbl = rospy.Publisher(controllerName+'/pose_updated',Empty)

def _camCallback(data):
    global image
    global img_pbl
    im = numpy.fromstring(data.data,dtype=numpy.uint8).reshape(15000,4)
    im.shape = (-1,4)
    image = im[:,[2,1,0,3]].reshape(100,150,4)
    emp = Empty()
    img_pbl.publish(emp)
    
def enableCamera():
    global isInitialized
    global isCameraEnabled
    if isInitialized and not isCameraEnabled:
        global controllerName
        global TIME_STEP
        global img_sb
        global img_pbl
        deviceName = 'kinectColor'
        sp = rospy.ServiceProxy(controllerName+'/'+deviceName+'/enable',sensor_enable)
        sp(TIME_STEP)
        img_sb = rospy.Subscriber(controllerName+'/'+deviceName+'/image',Image,_camCallback)
        img_pbl = rospy.Publisher(controllerName+'/'+deviceName+'/image_updated',Empty)
        isCameraEnabled = True
        
def disableCamera():
    global isCameraEnabled
    if isCameraEnabled:
        global img_sb
        img_sb.unregister()
        isCameraEnabled = False
        
def _usCallback(data):
    f = data.header.frame_id
    if f[-2] == 'o':
        so = int(f[-1])
    else:
        so = int(f[-2:])
    global distance
    distance[so] = data.range
    if so == 15:
        global so_pbl
        so_pbl.publish(Empty())
    
def enableSonar():
    global isInitialized
    global isSonarEnabled
    if isInitialized and not isSonarEnabled:
        global controllerName
        global TIME_STEP
        global so_sb
        global so_pbl
        so_sb = []
        for s in range(16):
            deviceName = 'so'+str(s)
            sp = rospy.ServiceProxy(controllerName+'/'+deviceName+'/enable',sensor_enable)
            sp(TIME_STEP)
            so_sb.append(rospy.Subscriber(controllerName+'/'+deviceName+'/value',Range,_usCallback))
        so_pbl = rospy.Publisher(controllerName+'/sonar_updated',Empty)
        isSonarEnabled = True
            
def disableSonar():
    global isSonarEnabled
    if isSonarEnabled:
        global so_sb
        for sb in so_sb:
            sb.unregister()
        isSonarEnabled = False
        
def _rngCallback(data):
    global range_image
    im = numpy.fromstring(data.data,dtype=numpy.uint8).reshape(15000,4)
    im.shape = (-1,4)
    range_image = im[:,[2,1,0,3]].reshape(100,150,4)

def enableRangeCamera():
    global isInitialized
    global isRangeCameraEnabled
    if isInitialized and not isRangeCameraEnabled:
        global controllerName
        global TIME_STEP
        global rng_sb
        deviceName = 'kinectRange'
        sp = rospy.ServiceProxy(controllerName+'/'+deviceName+'/enable',sensor_enable)
        sp(TIME_STEP)
        rng_sb = rospy.Subscriber(controllerName+'/'+deviceName+'/range_image',Image,_rngCallback)
        isRangeCameraEnabled = True
        
def disableRangeCamera():
    global isRangeCameraEnabled
    if isRangeCameraEnabled:
        global rng_sb
        rng_sb.unregister()
        isRangeCameraEnabled = False
